@@ -28,7 +28,14 @@ with this program.  If not, see <http://www.gnu.org/licenses/>.
 #include <gdk/gdkkeysyms.h>
 #include <gtk/gtk.h>
 
-#include "libindicator/indicator-object.h"
+#include <libindicator/indicator-object.h>
+
+/* For new style indicators */
+#if HAVE_INDICATOR_NG
+#include <libido/libido.h>
+#include <libindicator/indicator-ng.h>
+#endif
+
 #include "tomboykeybinder.h"
 
 static gchar * indicator_order[] = {
@@ -291,6 +298,11 @@ entry_added (IndicatorObject * io, IndicatorObjectEntry * entry, GtkWidget * men
 		gtk_hbox_new(FALSE, 3) : gtk_vbox_new(FALSE, 3);
 #endif
 
+#if GTK_CHECK_VERSION (3, 0, 0)
+	/* Allows indicators to receive mouse scroll event in GTK+3 */
+	gtk_widget_add_events(GTK_WIDGET(menuitem), GDK_SCROLL_MASK);
+#endif
+
 	g_object_set_data (G_OBJECT (menuitem), "indicator", io);
 	g_object_set_data (G_OBJECT (menuitem), "box", box);
 
@@ -505,29 +517,23 @@ update_accessible_desc(IndicatorObjectEntry * entry, GtkWidget * menuitem)
 	return;
 }
 
-
-static gboolean
-load_module (const gchar * name, GtkWidget * menubar)
+static void
+load_indicator (GtkWidget * menubar, IndicatorObject *io, const gchar *name)
 {
-	g_debug("Looking at Module: %s", name);
-	g_return_val_if_fail(name != NULL, FALSE);
-
-	if (!g_str_has_suffix(name, G_MODULE_SUFFIX)) {
-		return FALSE;
-	}
-
-	g_debug("Loading Module: %s", name);
-
-	/* Build the object for the module */
-	gchar * fullpath = g_build_filename(INDICATOR_DIR, name, NULL);
-	IndicatorObject * io = indicator_object_new_from_file(fullpath);
-	g_free(fullpath);
-
 	/* Set the environment it's in */
 	indicator_object_set_environment(io, (const GStrv)indicator_env);
 
 	/* Attach the 'name' to the object */
-	g_object_set_data(G_OBJECT(io), IO_DATA_ORDER_NUMBER, GINT_TO_POINTER(name2order(name)));
+#if HAVE_INDICATOR_NG
+	int pos = 5000 - indicator_object_get_position(io);
+	if (pos > 5000) {
+		pos = name2order(name);
+	}
+#else
+	int pos = name2order(name);
+#endif
+
+	g_object_set_data(G_OBJECT(io), IO_DATA_ORDER_NUMBER, GINT_TO_POINTER(pos));
 
 	/* Connect to its signals */
 	g_signal_connect(G_OBJECT(io), INDICATOR_OBJECT_SIGNAL_ENTRY_ADDED,   G_CALLBACK(entry_added),    menubar);
@@ -546,10 +552,129 @@ load_module (const gchar * name, GtkWidget * menubar)
 	}
 
 	g_list_free(entries);
+}
+
+static gboolean
+load_module (const gchar * name, GtkWidget * menubar)
+{
+	g_debug("Looking at Module: %s", name);
+	g_return_val_if_fail(name != NULL, FALSE);
+
+	if (!g_str_has_suffix(name, G_MODULE_SUFFIX)) {
+		return FALSE;
+	}
+
+	g_debug("Loading Module: %s", name);
+
+	/* Build the object for the module */
+	gchar * fullpath = g_build_filename(INDICATOR_DIR, name, NULL);
+	IndicatorObject * io = indicator_object_new_from_file(fullpath);
+	g_free(fullpath);
+
+	load_indicator(menubar, io, name);
 
 	return TRUE;
 }
 
+static void
+load_modules (GtkWidget *menubar, gint *indicators_loaded)
+{
+	if (g_file_test(INDICATOR_DIR, (G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR))) {
+		GDir * dir = g_dir_open(INDICATOR_DIR, 0, NULL);
+
+		const gchar * name;
+		gint count = 0;
+		while ((name = g_dir_read_name(dir)) != NULL) {
+#ifdef INDICATOR_APPLET_APPMENU
+			if (g_strcmp0(name, "libappmenu.so")) {
+				continue;
+			}
+#else
+			if (!g_strcmp0(name, "libappmenu.so")) {
+				continue;
+			}
+#endif
+#ifdef INDICATOR_APPLET
+			if (!g_strcmp0(name, "libme.so")) {
+				continue;
+			}
+			if (!g_strcmp0(name, "libdatetime.so")) {
+				continue;
+			}
+#endif
+			if (load_module(name, menubar)) {
+				count++;
+			}
+		}
+
+		*indicators_loaded += count;
+
+		g_dir_close (dir);
+	}
+}
+
+#if HAVE_INDICATOR_NG
+
+#define INDICATOR_SERVICE_DIR "/usr/share/unity/indicators"
+
+static void
+load_indicators_from_indicator_files (GtkWidget *menubar, gint *indicators_loaded)
+{
+	GDir *dir;
+	const gchar *name;
+	GError *error = NULL;
+
+	dir = g_dir_open (INDICATOR_SERVICE_DIR, 0, &error);
+
+	if (!dir) {
+		g_warning ("unable to open indicator service file directory: %s", error->message);
+		g_error_free (error);
+
+		return;
+	}
+
+	gint count = 0;
+	while ((name = g_dir_read_name (dir))) {
+		gchar *filename;
+		IndicatorNg *indicator;
+
+		filename = g_build_filename (INDICATOR_SERVICE_DIR, name, NULL);
+		indicator = indicator_ng_new_for_profile (filename, "desktop", &error);
+		g_free (filename);
+
+#ifdef INDICATOR_APPLET_APPMENU
+		if (g_strcmp0(name, "com.canonical.indicator.appmenu")) {
+			continue;
+		}
+#else
+		if (!g_strcmp0(name, "com.canonical.indicator.appmenu")) {
+			continue;
+		}
+#endif
+#ifdef INDICATOR_APPLET
+		if (!g_strcmp0(name, "com.canonical.indicator.me")) {
+			continue;
+		}
+		if (!g_strcmp0(name, "com.canonical.indicator.datetime")) {
+			continue;
+		}
+#endif
+
+		if (indicator) {
+			load_indicator(menubar, INDICATOR_OBJECT (indicator), name);
+			count++;
+		}else{
+			g_warning ("unable to load '%s': %s", name, error->message);
+			g_clear_error (&error);
+		}
+	}
+
+	*indicators_loaded += count;
+
+	g_dir_close (dir);
+}
+#endif  /* HAVE_INDICATOR_NG */
+
 static void
 hotkey_filter (char * keystring G_GNUC_UNUSED, gpointer data)
 {
@@ -797,6 +922,10 @@ static gboolean
 applet_fill_cb (MatePanelApplet * applet, const gchar * iid G_GNUC_UNUSED,
                 gpointer data G_GNUC_UNUSED)
 {
+#if HAVE_INDICATOR_NG
+	ido_init();
+#endif
+
 	static const GtkActionEntry menu_actions[] = {
 		{"About", GTK_STOCK_ABOUT, N_("_About"), NULL, NULL, G_CALLBACK(about_cb)}
 	};
@@ -903,35 +1032,10 @@ applet_fill_cb (MatePanelApplet * applet, const gchar * iid G_GNUC_UNUSED,
 	/* Add in filter func */
 	tomboy_keybinder_bind(hotkey_keycode, hotkey_filter, menubar);
 
-	/* load 'em */
-	if (g_file_test(INDICATOR_DIR, (G_FILE_TEST_EXISTS | G_FILE_TEST_IS_DIR))) {
-		GDir * dir = g_dir_open(INDICATOR_DIR, 0, NULL);
-
-		const gchar * name;
-		while ((name = g_dir_read_name(dir)) != NULL) {
-#ifdef INDICATOR_APPLET_APPMENU
-			if (g_strcmp0(name, "libappmenu.so")) {
-				continue;
-			}
-#else
-			if (!g_strcmp0(name, "libappmenu.so")) {
-				continue;
-			}
+	load_modules(menubar, &indicators_loaded);
+#if HAVE_INDICATOR_NG
+	load_indicators_from_indicator_files(menubar, &indicators_loaded);
 #endif
-#ifdef INDICATOR_APPLET
-			if (!g_strcmp0(name, "libme.so")) {
-				continue;
-			}
-			if (!g_strcmp0(name, "libdatetime.so")) {
-				continue;
-			}
-#endif
-			if (load_module(name, menubar)) {
-				indicators_loaded++;
-			}
-		}
-		g_dir_close (dir);
-	}
 
 	if (indicators_loaded == 0) {
 		/* A label to allow for click through */
